(require 'bongo)
(require 'org)

;;; customization

(defgroup transcribe nil
  "Transcribe voice memo recordings"
  :group 'outlines
  :group 'hypermedia
  :group 'bongo
  :group 'calendar)

;;; QuallComm PureVoice

(defcustom qcp-date-regexp
  (concat "^\\([[:digit:]]\\{2\\}\\)" ;; month
          "\\([[:digit:]]\\{2\\}\\)" ;; day
          "\\([[:digit:]]\\{2\\}\\)" ;; year
          "\\([[:digit:]]\\{2\\}\\)" ;; hour
          "\\([[:digit:]]\\{2\\}\\)" ;; minute
          "[[:ascii:]]?$")
  "Regexp used to identify and extract date and time stamp from the
  base name of a filename of a QualComm PureVoice memo."
  :type 'regexp
  :group 'transcribe)

(defun qcp-date-to-time (file-name)
  (let ((name (file-name-sans-extension
               (file-name-nondirectory file-name))))
    (if (string-match qcp-date-regexp name)
        (encode-time 0
                     (string-to-number (match-string 5 name))
                     (string-to-number (match-string 4 name))
                     (string-to-number (match-string 2 name))
                     (string-to-number (match-string 1 name))
                     (+ (* (/ (nth 5 (decode-time)) 1000) 1000)
                        (string-to-number (match-string 3 name)))))))

(defun qcp-current-track-time ()
  (with-bongo-playlist-buffer
    (if (marker-position bongo-current-track-marker)
         (qcp-date-to-time (bongo-line-file-name
                            bongo-current-track-marker)))))

(defun qcp-org-current-time-stamp ()
  (with-bongo-playlist-buffer
    (if (numberp (marker-position bongo-current-track-marker))
        (format-time-string (org-time-stamp-format 'long 'inactive)
                            (qcp-current-track-time)))))

(defun qcp-insert-current-time-stamp ()
  (interactive)
  (org-insert-time-stamp (qcp-current-track-time) t t))

(defun pvconv (infile)
  "Use pvconv to convert INFILE to a file next to infile.  The
  absolute path to the converted file is returned."
  (call-process "pvconv" nil "*pvconv*" nil infile)
  (concat
   (file-name-sans-extension infile)
   (if (equal (file-name-extension infile t) ".qcp")
       ".wav" ".qcp")))

(define-bongo-backend qcp
  :pretty-name "PureVoice"
  :constructor 'bongo-start-qcp-player
  :matcher '(local-file "qcp"))

(defun pvconv-cleanup-player (player)
  "Remove the file converted by pvconv.  Intended for use as a bongo
  player hook such as bongo-player-stopped-functions."
  (if (bongo-player-get player 'pvconv-file-name)
      (let ((file-name (bongo-player-get player 'file-name)))
        (if (file-regular-p file-name)
            (delete-file file-name)))))

(add-hook 'bongo-player-stopped-functions 'pvconv-cleanup-player)

(defun bongo-start-qcp-player (file-name &optional extra-arguments)
  (let* ((wav-file-name (pvconv file-name))
         (player (bongo-play-file wav-file-name)))
    (bongo-player-put player 'pvconv-file-name file-name)
    player))

;;; Bongo

(defun transcribe/bongo-current-file ()
  (with-bongo-playlist-buffer
    (and (marker-position bongo-current-track-marker)
         (bongo-local-file-track-line-p bongo-current-track-marker)
         (bongo-line-file-name bongo-current-track-marker))))

(defun transcribe/bongo-rename-line ()
  "Call bongo-rename-line on the line for the current track.  Use
  read-file-name if the line is a local file for a better input
  interface.  The track is replayed after moving."
  (interactive)
  (with-bongo-playlist-buffer
    (if (bongo-local-file-track-line-p)
        (let* ((file (bongo-line-file-name
                      bongo-current-track-marker))
               (dir (file-name-directory file))
               (base (file-name-nondirectory file))
               (new (read-file-name
                         (format "Move %s to: " file) dir base))
               (new (if (file-directory-p new)
                        (expand-file-name base new) new)))
          (bongo-rename-local-file-track
           new bongo-current-track-marker))
      (call-interactively 'bongo-rename-line))
    (bongo-play-previous)))

(global-set-key "\C-cvr" 'transcribe/bongo-rename-line)
(global-set-key "\C-cvs" 'bongo-seek)

;;; Org-mode

;; Modify the org-log-note-headings customize variable to include our
;; values
(defcustom org-log-note-headings
  '((done .  "CLOSING NOTE %t")
    (state . "State %-12s %t")
    (note .  "Note taken on %t")
    (clock-out . "")
    (transcribe . "[[file://%f][Recorded]] on %r"))
  "Headings for notes added to entries.
The value is an alist, with the car being a symbol indicating the note
context, and the cdr is the heading to be used.  The heading may also be the
empty string.
%t in the heading will be replaced by a time stamp.
%s will be replaced by the new TODO state, in double quotes.
%u will be replaced by the user name.
%U will be replaced by the full user name.
%r will be replaced by the recording date.
%f will be replaced by the recording file."
  :group  'org-todo
  :group  'org-progress
  :type '(list :greedy t
	  (cons (const :tag "Heading when closing an item" done) string)
	  (cons (const :tag
		       "Heading when changing todo state (todo sequence only)"
		       state) string)
	  (cons (const :tag "Heading when just taking a note" note) string)
	  (cons (const :tag "Heading when clocking out" clock-out) string)
          (cons (const :tag "Heading when transcribing" note) string)))

;; Add support for transcription notes to the org-add-log-note
;; function

(defun org-add-log-note (&optional purpose)
  "Pop up a window for taking a note, and add this note later at point."
  (remove-hook 'post-command-hook 'org-add-log-note)
  (setq org-log-note-window-configuration (current-window-configuration))
  (delete-other-windows)
  (move-marker org-log-note-return-to (point))
  (switch-to-buffer (marker-buffer org-log-note-marker))
  (goto-char org-log-note-marker)
  (org-switch-to-buffer-other-window "*Org Note*")
  (erase-buffer)
  (if (memq org-log-note-how '(time state))
      (org-store-log-note)
    (let ((org-inhibit-startup t)) (org-mode))
    (insert (format "# Insert note for %s.
# Finish with C-c C-c, or cancel with C-c C-k.\n\n"
		    (cond
		     ((eq org-log-note-purpose 'clock-out) "stopped clock")
		     ((eq org-log-note-purpose 'done)  "closed todo item")
		     ((eq org-log-note-purpose 'state)
		      (format "state change to \"%s\"" org-log-note-state))
		     ((memq org-log-note-purpose '(note transcribe))
		      "this entry")
		     (t (error "This should not happen")))))
    (org-set-local 'org-finish-function 'org-store-log-note)))

(defun org-store-log-note ()
  "Finish taking a log note, and insert it to where it belongs."
  (let ((txt (buffer-string))
	(note (cdr (assq org-log-note-purpose org-log-note-headings)))
	lines ind)
    (kill-buffer (current-buffer))
    (while (string-match "\\`#.*\n[ \t\n]*" txt)
      (setq txt (replace-match "" t t txt)))
    (if (string-match "\\s-+\\'" txt)
	(setq txt (replace-match "" t t txt)))
    (setq lines (org-split-string txt "\n"))
    (when (and note (string-match "\\S-" note))
      (setq note
	    (org-replace-escapes
	     note
	     (list (cons "%u" (user-login-name))
		   (cons "%U" user-full-name)
		   (cons "%t" (format-time-string
			       (org-time-stamp-format 'long 'inactive)
			       (current-time)))
		   (cons "%s" (if org-log-note-state
				  (concat "\"" org-log-note-state "\"")
				""))
		   (cons "%r" (qcp-org-current-time-stamp))
		   (cons "%f" (transcribe/bongo-current-file)))))
      (if lines (setq note (concat note " \\\\")))
      (push note lines))
    (when (or current-prefix-arg org-note-abort) (setq lines nil))
    (when lines
      (save-excursion
	(set-buffer (marker-buffer org-log-note-marker))
	(save-excursion
	  (goto-char org-log-note-marker)
	  (move-marker org-log-note-marker nil)
	  (end-of-line 1)
	  (if (not (bolp)) (let ((inhibit-read-only t)) (insert "\n")))
	  (indent-relative nil)
	  (insert "- " (pop lines))
	  (org-indent-line-function)
	  (beginning-of-line 1)
	  (looking-at "[ \t]*")
	  (setq ind (concat (match-string 0) "  "))
	  (end-of-line 1)
	  (while lines (insert "\n" ind (pop lines)))))))
  (set-window-configuration org-log-note-window-configuration)
  (with-current-buffer (marker-buffer org-log-note-return-to)
    (goto-char org-log-note-return-to))
  (move-marker org-log-note-return-to nil)
  (and org-log-post-message (message "%s" org-log-post-message)))

(defun org-add-transcription ()
  "Add a transcription to the current entry.
This is done in the same way as adding a state change note."
  (interactive)
  (org-add-log-setup 'transcribe nil t nil))

(global-set-key "\C-cvz" 'org-add-transcription)

(defun transcribe-move-and-store-link (&optional file newname)
  "Move the file or the currently playing bongo file also storing a
  link to the new location."
  (interactive)
  (with-bongo-playlist-buffer
    (let* ((basename (file-name-nondirectory file))
           (basename-wo-ext (file-name-sans-extension basename))
           (desc
            (if (string-match transcribe-file-regexp basename-wo-ext)
                (format-time-string 
                 (concat "[" 
                         (substring (cdr org-time-stamp-formats) 1 -1) 
                         "]") 
                 (org-read-date 
                  nil t (format "%s-%s-%s %s:%s"
                                (match-string 3 basename-wo-ext)
                                (match-string 1 basename-wo-ext)
                                (match-string 2 basename-wo-ext)
                                (match-string 4 basename-wo-ext)
                                (match-string 5 basename-wo-ext))))
              basename-wo-ext)))
      (setq org-stored-links
            (cons (list (concat "file://" newname) desc)
                  org-stored-links)))))

;;;; Finish up
(provide 'transcribe)
